"""Authlib implementation of DIDComm crypto."""

import hashlib
import json
from typing import Mapping, Optional, Sequence, Tuple, Union

from pydid import VerificationMethod

from didcomm_messaging.crypto.base import (
    CryptoService,
    CryptoServiceError,
    PublicKey,
    SecretKey,
)
from didcomm_messaging.multiformats import multibase, multicodec
from didcomm_messaging.multiformats.multibase import Base64UrlEncoder


try:
    from authlib.jose import JsonWebEncryption, JsonWebKey
    from authlib.jose.rfc7517 import AsymmetricKey
    from authlib.jose.drafts import register_jwe_draft

    register_jwe_draft(JsonWebEncryption)
except ImportError:
    raise ImportError("Authlib backend requires the 'authlib' extra to be installed")

b64url = Base64UrlEncoder()


class AuthlibKey(PublicKey):
    """Authlib implementation of DIDComm PublicKey."""

    kty_crv_to_codec: Mapping[Tuple[str, Optional[str]], str] = {
        ("OKP", "Ed25519"): "ed25519-pub",
        ("OKP", "X25519"): "x25519-pub",
    }
    codec_to_kty_crv: Mapping[str, Tuple[str, str]] = {
        "ed25519-pub": ("OKP", "Ed25519"),
        "x25519-pub": ("OKP", "X25519"),
    }

    def __init__(self, key: AsymmetricKey, kid: str):
        """Initialize the AuthlibKey."""
        self.key = key
        self._kid = kid
        self._multikey = self.key_to_multikey(key)

    @property
    def kid(self) -> str:
        """Return the key ID."""
        return self._kid

    @property
    def multikey(self) -> str:
        """Return the key in multikey format."""
        return self._multikey

    @classmethod
    def key_to_multikey(cls, key: AsymmetricKey) -> str:
        """Convert an Authlib key to a multikey."""
        jwk = key.as_dict(is_private=False)
        codec = cls.kty_crv_to_codec.get((jwk["kty"], jwk.get("crv")))

        if not codec:
            raise ValueError("Unsupported key type")

        key_bytes = b64url.decode(jwk["x"])
        return multibase.encode(
            multicodec.wrap(multicodec.multicodec(codec), key_bytes), "base58btc"
        )

    @classmethod
    def multikey_to_key(cls, multikey: str) -> AsymmetricKey:
        """Convert a multikey to an Authlib key."""
        decoded = multibase.decode(multikey)
        codec, key = multicodec.unwrap(decoded)
        try:
            kty, crv = cls.codec_to_kty_crv[codec.name]
        except KeyError:
            raise ValueError(f"Unsupported key type: {codec.name}")

        jwk = {"kty": kty, "crv": crv, "x": b64url.encode(key)}

        try:
            return JsonWebKey.import_key(jwk)
        except Exception as err:
            raise ValueError("Invalid key") from err

    @classmethod
    def from_verification_method(cls, vm: VerificationMethod) -> "AuthlibKey":
        """Return a PublicKey from a verification method."""
        # TODO Reduce code duplication between this and Askar
        if not vm.id.did:
            kid = vm.id.as_absolute(vm.controller)
        else:
            kid = vm.id

        if vm.type == "Multikey":
            multikey = vm.public_key_multibase
            if not multikey:
                raise ValueError("Multikey verification method missing key")

            key = cls.multikey_to_key(multikey)
            return cls(key, kid)

        codec = cls.type_to_codec.get(vm.type)
        if not codec:
            raise ValueError("Unsupported verification method type: {vm_type}")

        try:
            kty, crv = cls.codec_to_kty_crv[codec]
        except KeyError:
            raise ValueError("Unsupported verification method type")

        key_bytes = cls.key_bytes_from_verification_method(vm)
        jwk = {"kty": kty, "crv": crv, "x": b64url.encode(key_bytes)}

        key = JsonWebKey.import_key(jwk)
        return cls(key, kid)


class AuthlibSecretKey(SecretKey):
    """Authlib implementation of SecretKey."""

    def __init__(self, key: AsymmetricKey, kid: str):
        """Initialize the AuthlibSecretKey."""
        self.key = key
        self._kid = kid

    @property
    def kid(self) -> str:
        """Return the key ID."""
        return self._kid


class AuthlibCryptoService(CryptoService[AuthlibKey, AuthlibSecretKey]):
    """Authlib implementation of CryptoService."""

    @classmethod
    def verification_method_to_public_key(cls, vm: VerificationMethod) -> AuthlibKey:
        """Return a PublicKey from a verification method."""
        return AuthlibKey.from_verification_method(vm)

    def _build_header_ecdh_1pu(
        self, to: Sequence[AuthlibKey], frm: AuthlibSecretKey, alg: str, enc: str
    ):
        skid = frm.kid
        kids = [to_key.kid for to_key in to]

        apu = b64url.encode(skid.encode())
        apv = b64url.encode(hashlib.sha256((".".join(sorted(kids))).encode()).digest())
        protected = {
            "typ": "application/didcomm-encrypted+json",
            "alg": alg,
            "enc": enc,
            "apu": apu,
            "apv": apv,
            "skid": skid,
        }
        recipients = [{"header": {"kid": kid}} for kid in kids]
        return {"protected": protected, "recipients": recipients}

    def _build_header_ecdh_es(self, to: Sequence[AuthlibKey], alg: str, enc: str):
        kids = [to_key.kid for to_key in to]

        apv = b64url.encode(hashlib.sha256((".".join(sorted(kids))).encode()).digest())
        protected = {
            "typ": "application/didcomm-encrypted+json",
            "alg": alg,
            "enc": enc,
            "apv": apv,
        }
        recipients = [{"header": {"kid": kid}} for kid in kids]
        return {"protected": protected, "recipients": recipients}

    async def ecdh_es_encrypt(
        self, to_keys: Sequence[AuthlibKey], message: bytes
    ) -> bytes:
        """Encrypt a message using ECDH-ES."""
        header = self._build_header_ecdh_es(to_keys, "ECDH-ES+A256KW", "XC20P")
        jwe = JsonWebEncryption()
        res = jwe.serialize_json(header, message, [value.key for value in to_keys])
        return json.dumps(res).encode()

    async def ecdh_es_decrypt(
        self, enc_message: Union[str, bytes], recip_key: AuthlibSecretKey
    ) -> bytes:
        """Decrypt a message using ECDH-ES."""
        try:
            jwe = JsonWebEncryption()
            res = jwe.deserialize_json(enc_message, recip_key.key)
        except Exception as err:
            raise CryptoServiceError("Invalid JWE") from err

        return res["payload"]

    async def ecdh_1pu_encrypt(
        self,
        to_keys: Sequence[AuthlibKey],
        sender_key: AuthlibSecretKey,
        message: bytes,
    ) -> bytes:
        """Encrypt a message using ECDH-1PU."""
        header = self._build_header_ecdh_1pu(
            to_keys, sender_key, "ECDH-1PU+A256KW", "A256CBC-HS512"
        )
        jwe = JsonWebEncryption()
        res = jwe.serialize_json(
            header, message, [value.key for value in to_keys], sender_key=sender_key.key
        )
        return json.dumps(res).encode()

    async def ecdh_1pu_decrypt(
        self,
        enc_message: Union[str, bytes],
        recip_key: AuthlibSecretKey,
        sender_key: AuthlibKey,
    ) -> bytes:
        """Decrypt a message using ECDH-1PU."""
        try:
            jwe = JsonWebEncryption()
            res = jwe.deserialize_json(
                enc_message, recip_key.key, sender_key=sender_key.key
            )
        except Exception as err:
            raise CryptoServiceError("Invalid JWE") from err

        return res["payload"]



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--runexternal",
        action="store_true",
        default=False,
        help="run tests that make external requests",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "external_fetch: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runexternal"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_external = pytest.mark.skip(reason="need --runexternal option to run")
    for item in items:
        if "external_fetch" in item.keywords:
            item.add_marker(skip_external)



"""PackagingService interface."""

from dataclasses import dataclass
import hashlib
from typing import Generic, Literal, Optional, Sequence, Tuple, Union

from pydid import DIDUrl, VerificationMethod
from didcomm_messaging.crypto import P, S, CryptoService, SecretsManager
from didcomm_messaging.crypto.jwe import JweEnvelope, b64url, from_b64url
from didcomm_messaging.resolver import DIDResolver


@dataclass
class PackedMessageMetadata(Generic[S]):
    """Unpack result."""

    wrapper: JweEnvelope
    method: Literal["ECDH-ES", "ECDH-1PU"]
    recip_key: S
    sender_kid: Optional[str]


class PackagingServiceError(Exception):
    """Represents an error from the DIDComm Messaging interface."""


class PackagingService(Generic[P, S]):
    """DIDComm Messaging interface."""

    async def extract_packed_message_metadata(  # noqa: C901
        self, enc_message: Union[str, bytes], secrets: SecretsManager[S]
    ) -> PackedMessageMetadata:
        """Extract metadata from a packed DIDComm message."""
        try:
            wrapper = JweEnvelope.from_json(enc_message)
        except ValueError:
            raise PackagingServiceError("Invalid packed message")

        alg = wrapper.protected.get("alg")
        if not alg:
            raise PackagingServiceError("Missing alg header")

        method = next((m for m in ("ECDH-1PU", "ECDH-ES") if m in alg), None)
        if not method:
            raise PackagingServiceError(
                f"Unsupported DIDComm encryption algorithm: {alg}"
            )

        sender_kid = None
        recip_key = None
        for kid in wrapper.recipient_key_ids:
            recip_key = await secrets.get_secret_by_kid(kid)
            if recip_key:
                break

        if not recip_key:
            raise PackagingServiceError("No recognized recipient key")

        expected_apv = b64url(
            hashlib.sha256((".".join(wrapper.recipient_key_ids)).encode()).digest()
        )
        apv = wrapper.protected.get("apv")
        if not apv:
            raise PackagingServiceError("Missing apv header")
        if apv != expected_apv:
            raise PackagingServiceError("Invalid apv value")

        if method == "ECDH-1PU":
            sender_kid_apu = None
            apu = wrapper.protected.get("apu")
            if not apu:
                raise PackagingServiceError("Missing apu header")

            try:
                sender_kid_apu = from_b64url(apu).decode("utf-8")
            except (UnicodeDecodeError, ValueError):
                raise PackagingServiceError("Invalid apu value")

            sender_kid = wrapper.protected.get("skid") or sender_kid_apu
            if sender_kid != sender_kid_apu:
                raise PackagingServiceError("Mismatch between skid and apu")
            if not sender_kid:
                raise PackagingServiceError("Sender key ID not provided")

        return PackedMessageMetadata(wrapper, method, recip_key, sender_kid)

    async def unpack(
        self,
        crypto: CryptoService[P, S],
        resolver: DIDResolver,
        secrets: SecretsManager[S],
        enc_message: Union[str, bytes],
    ) -> Tuple[bytes, PackedMessageMetadata]:
        """Unpack a DIDComm message."""
        metadata = await self.extract_packed_message_metadata(enc_message, secrets)

        if metadata.method == "ECDH-ES":
            return (
                await crypto.ecdh_es_decrypt(enc_message, metadata.recip_key),
                metadata,
            )

        if not metadata.sender_kid:
            raise PackagingServiceError("Missing sender key ID")

        sender_vm = await resolver.resolve_and_dereference_verification_method(
            metadata.sender_kid
        )
        sender_key = crypto.verification_method_to_public_key(sender_vm)

        return (
            await crypto.ecdh_1pu_decrypt(enc_message, metadata.recip_key, sender_key),
            metadata,
        )

    async def recip_for_kid_or_default_for_did(
        self, crypto: CryptoService[P, S], resolver: DIDResolver, kid_or_did: str
    ) -> P:
        """Resolve a verification method for a kid or return default recip."""
        if "#" in kid_or_did:
            vm = await resolver.resolve_and_dereference_verification_method(kid_or_did)
        else:
            doc = await resolver.resolve_and_parse(kid_or_did)
            if not doc.key_agreement:
                raise PackagingServiceError(
                    "No key agreement methods found; cannot determine recipient"
                )

            default = doc.key_agreement[0]
            if isinstance(default, DIDUrl):
                vm = doc.dereference(default)
                if not isinstance(vm, VerificationMethod):
                    raise PackagingServiceError(
                        f"Expected verification method, found: {type(vm)}"
                    )
            else:
                vm = default

        return crypto.verification_method_to_public_key(vm)

    async def default_sender_kid_for_did(self, resolver: DIDResolver, did: str) -> str:
        """Determine the kid of the default sender key for a DID."""
        if "#" in did:
            return did

        doc = await resolver.resolve_and_parse(did)
        if not doc.key_agreement:
            raise PackagingServiceError(
                "No key agreement methods found; cannot determine recipient"
            )

        default = doc.key_agreement[0]
        if isinstance(default, DIDUrl):
            vm = doc.dereference(default)
            if not isinstance(vm, VerificationMethod):
                raise PackagingServiceError(
                    f"Expected verification method, found: {type(vm)}"
                )
        else:
            vm = default

        if not vm.id.did:
            return vm.id.as_absolute(vm.controller)
        return vm.id

    async def pack(
        self,
        crypto: CryptoService[P, S],
        resolver: DIDResolver,
        secrets: SecretsManager[S],
        message: bytes,
        to: Sequence[str],
        frm: Optional[str] = None,
        **options,
    ):
        """Pack a DIDComm message."""
        recip_keys = [
            await self.recip_for_kid_or_default_for_did(crypto, resolver, kid)
            for kid in to
        ]
        sender_kid = await self.default_sender_kid_for_did(resolver, frm) if frm else None
        sender_key = await secrets.get_secret_by_kid(sender_kid) if sender_kid else None
        if frm and not sender_key:
            raise PackagingServiceError("No sender key found")

        if sender_key:
            return await crypto.ecdh_1pu_encrypt(recip_keys, sender_key, message)
        else:
            return await crypto.ecdh_es_encrypt(recip_keys, message)



import pytest

from didcomm_messaging.resolver import (
    DIDMethodNotSupported,
    DIDResolver,
    PrefixResolver,
)


class TestResolver(DIDResolver):
    async def is_resolvable(self, did: str) -> bool:
        return True

    async def resolve(self, did: str) -> dict:
        return {"did": did}


@pytest.fixture(scope="session")
def test_resolver():
    yield TestResolver()


@pytest.mark.asyncio
async def test_prefix_resolver(test_resolver: DIDResolver):
    did = "did:test:example_did"

    resolver = PrefixResolver(resolvers={"did:test:": test_resolver})

    doc = await resolver.resolve(did)
    assert doc["did"] == did

    with pytest.raises(DIDMethodNotSupported):
        await resolver.resolve("This won't work")



"""DIDComm Messaging Cryptography and Secrets Interfaces."""

from .base import CryptoService, SecretsManager, PublicKey, SecretKey, P, S

__all__ = ["CryptoService", "SecretsManager", "PublicKey", "SecretKey", "P", "S"]



"""DID Resolver."""

from abc import ABC, abstractmethod
from typing import Dict
from pydid import DIDDocument, DIDUrl, Resource, VerificationMethod


class DIDResolutionError(Exception):
    """Represents an error from a DID Resolver."""


class DIDNotFound(DIDResolutionError):
    """Represents a DID not found error."""


class DIDMethodNotSupported(DIDResolutionError):
    """Represents a DID method not supported error."""


class DIDResolver(ABC):
    """DID Resolver interface."""

    @abstractmethod
    async def resolve(self, did: str) -> dict:
        """Resolve a DID."""

    @abstractmethod
    async def is_resolvable(self, did: str) -> bool:
        """Check to see if a DID is resolvable."""

    async def resolve_and_parse(self, did: str) -> DIDDocument:
        """Resolve a DID and parse the DID document."""
        doc = await self.resolve(did)
        return DIDDocument.deserialize(doc)

    async def resolve_and_dereference(self, did_url: str) -> Resource:
        """Resolve a DID URL and dereference the identifier."""
        url = DIDUrl.parse(did_url)
        if not url.did:
            raise DIDResolutionError("Invalid DID URL; must be absolute")

        doc = await self.resolve_and_parse(url.did)
        return doc.dereference(url)

    async def resolve_and_dereference_verification_method(
        self, did_url: str
    ) -> VerificationMethod:
        """Resolve a DID URL and dereference the identifier."""
        resource = await self.resolve_and_dereference(did_url)
        if not isinstance(resource, VerificationMethod):
            raise DIDResolutionError("Resource is not a verification method")

        return resource


class PrefixResolver(DIDResolver):
    """DID Resolver delegates to sub-resolvers by DID prefix."""

    def __init__(self, resolvers: Dict[str, DIDResolver]):
        """Initialize the resolver."""
        self.resolvers = resolvers

    async def is_resolvable(self, did: str) -> bool:
        """Check to see if a DID is resolvable."""
        for prefix, resolver in self.resolvers.items():
            if did.startswith(prefix):
                return await resolver.is_resolvable(did)
        return False

    async def resolve(self, did: str) -> dict:
        """Resolve a DID."""
        for prefix, resolver in self.resolvers.items():
            if did.startswith(prefix):
                return await resolver.resolve(did)

        raise DIDMethodNotSupported(f"No resolver found for DID {did}")



import pytest
from didcomm_messaging.crypto import SecretKey

from didcomm_messaging.crypto.backend.basic import InMemorySecretsManager


class MockSecretKey(SecretKey):
    def __init__(self, kid) -> None:
        self._kid = kid

    @property
    def kid(self):
        return self._kid


@pytest.fixture()
def in_memory_secrets_manager():
    yield InMemorySecretsManager()


@pytest.fixture()
def secret():
    kid = "did:example:alice#key-1"
    key = MockSecretKey(kid=kid)
    yield key


@pytest.mark.asyncio
async def test_in_memory_secrets(
    in_memory_secrets_manager: InMemorySecretsManager, secret: MockSecretKey
):
    await in_memory_secrets_manager.add_secret(secret)

    assert await in_memory_secrets_manager.get_secret_by_kid(secret.kid) == secret



"""did:web resolver.

Resolve did:web style dids to a did document. did:web spec:
https://w3c-ccg.github.io/did-method-web/
"""

from . import DIDResolver, DIDNotFound, DIDResolutionError
from pydid import DID
from urllib.parse import urlparse
from datetime import datetime, timedelta
import urllib.request as url_request
import re
import json
import urllib

domain_regex = (
    r"((?!-))(xn--)?[a-z0-9][a-z0-9-_]{0,61}[a-z0-9]{0,1}"
    r"\.(xn--)?([a-z0-9\._-]{1,61}|[a-z0-9-]{1,30})"
    r"(%3[aA]\d+)?"  # Port
    r"(:[a-zA-Z]+)*"  # Path
)
did_web_pattern = re.compile(rf"^did:web:{domain_regex}$")
cache = {}
TIME_TO_CACHE = 1800  # 30 minutes


class DIDWeb(DIDResolver):
    """Utility functions for building and interacting with did:web."""

    async def resolve(self, did: str) -> dict:
        """Resolve a did:web to a did document via http request."""

        # Check to see if we've seen the did recently
        if did in cache:
            if cache[did]["timestamp"] > datetime.now() + timedelta(
                seconds=-TIME_TO_CACHE
            ):
                return cache[did]["doc"]
            else:
                del cache[did]

        uri = DIDWeb._did_to_uri(did)
        headers = {
            "User-Agent": "DIDCommRelay/1.0",
        }
        request = url_request.Request(url=uri, method="GET", headers=headers)
        try:
            with url_request.urlopen(request) as response:
                doc = json.loads(response.read().decode())
                cache[did] = {
                    "timestamp": datetime.now(),
                    "doc": doc,
                }
                return doc
        except urllib.error.HTTPError as e:
            if e.code == 404:
                raise DIDNotFound(
                    f"The did:web {did} returned a 404 not found while resolving"
                )
            else:
                raise DIDResolutionError(
                    f"Unknown server error ({e.code}) while resolving did:web: {did}"
                )
        except json.decoder.JSONDecodeError as e:
            msg = str(e)
            raise DIDNotFound(f"The did:web {did} returned invalid JSON {msg}")
        except Exception as e:
            raise DIDResolutionError("Failed to fetch did document") from e

    @staticmethod
    def _did_to_uri(did: str) -> str:
        # Split the did by it's segments
        did_segments = did.split(":")

        # Get the hostname & port
        hostname = did_segments[2].lower()
        hostname = hostname.replace("%3a", ":")

        # Resolve the path portion of the DID, if there is no path, default to
        # a .well-known address
        path = ".well-known"
        if len(did_segments) > 3:
            path = "/".join(did_segments[3:])

        # Assemble the URI
        did_uri = f"https://{hostname}/{path}/did.json"

        return did_uri

    async def is_resolvable(self, did: str) -> bool:
        """Determine if the did is a valid did:web did that can be resolved."""
        if DID.is_valid(did) and did_web_pattern.match(did):
            return True
        return False

    @staticmethod
    def did_from_url(url: str) -> DID:
        """Convert a URL into a did:web did."""

        # Make sure that the URL starts with a scheme
        if not url.startswith("http"):
            url = f"https://{url}"

        # Parse it out to we can grab pieces
        parsed_url = urlparse(url)

        # Assemble the domain portion of the DID
        did = "did:web:%s" % parsed_url.netloc.replace(":", "%3A")

        # Cleanup the path
        path = parsed_url.path.replace(".well-known/did.json", "")
        path = path.replace("/did.json", "")

        # Add the path portion of the did
        if len(path) > 1:
            did += path.replace("/", ":")
        return did



import pytest


from didcomm_messaging.resolver.web import DIDWeb

DIDWEB = "did:web:example.com"
DIDWEB_URI = "https://example.com/.well-known/did.json"
DIDWEB_COMPLEX = "did:web:example.com%3A4443:DIDs:alice:relay"
DIDWEB_COMPLEX_URI = "https://example.com:4443/DIDs/alice/relay/did.json"


@pytest.mark.asyncio
async def test_didweb_from_didurl_domain():
    did = DIDWeb.did_from_url("example.com")
    assert did
    assert did == DIDWEB


@pytest.mark.asyncio
async def test_didweb_from_didurl_schema_and_domain():
    did = DIDWeb.did_from_url("https://example.com")
    assert did
    assert did == DIDWEB


@pytest.mark.asyncio
async def test_didweb_from_didurl_schema_and_domain_slash():
    did = DIDWeb.did_from_url("https://example.com/")
    assert did
    assert did == DIDWEB


@pytest.mark.asyncio
async def test_didweb_from_didurl_schema_and_domain_path():
    did = DIDWeb.did_from_url("https://example.com/did.json")
    assert did
    assert did == DIDWEB


@pytest.mark.asyncio
async def test_didweb_from_didurl_schema_and_domain_wellknown():
    did = DIDWeb.did_from_url("https://example.com/.well-known/did.json")
    assert did
    assert did == DIDWEB


@pytest.mark.asyncio
async def test_didweb_from_didurl_schema_and_domain_port_wellknown():
    did = DIDWeb.did_from_url("https://example.com:443/.well-known/did.json")
    assert did
    assert did == DIDWEB + "%3A443"


@pytest.mark.asyncio
async def test_didweb_from_didurl_schema_and_complex_domain_path():
    did = DIDWeb.did_from_url("https://example.com:4443/DIDs/alice/relay/did.json")
    assert did
    assert did == DIDWEB_COMPLEX


@pytest.mark.asyncio
async def test_didweb_to_url():
    uri = DIDWeb._did_to_uri(DIDWEB)
    assert uri
    assert uri == DIDWEB_URI


@pytest.mark.asyncio
async def test_didweb_to_url_complex():
    uri = DIDWeb._did_to_uri(DIDWEB_COMPLEX)
    assert uri
    assert uri == DIDWEB_COMPLEX_URI


@pytest.mark.asyncio
async def test_didweb_is_resolvable():
    resolver = DIDWeb()
    resolvable = await resolver.is_resolvable(DIDWEB)
    assert resolvable
    resolvable_complex = await resolver.is_resolvable(DIDWEB_COMPLEX)
    assert resolvable_complex


@pytest.mark.external_fetch
@pytest.mark.asyncio
async def test_didweb_fetch():
    did_web = "did:web:colton.wolkins.net"
    resolver = DIDWeb()
    uri = await resolver.resolve(did_web)
    print(uri)
    assert uri
    assert isinstance(uri, dict)


@pytest.mark.external_fetch
@pytest.mark.asyncio
async def test_didweb_double_fetch():
    did_web = "did:web:colton.wolkins.net"
    resolver = DIDWeb()
    uri = await resolver.resolve(did_web)
    print(uri)
    assert uri
    assert isinstance(uri, dict)
    uri = await resolver.resolve(did_web)
    assert uri
    assert isinstance(uri, dict)



"""Test compabibility between Askar and Authlib."""

import json
from aries_askar import Key, KeyAlg
from authlib.jose import OKPKey
import pytest

from didcomm_messaging.crypto.backend.askar import (
    AskarCryptoService,
    AskarKey,
    AskarSecretKey,
)
from didcomm_messaging.crypto.backend.authlib import (
    AuthlibCryptoService,
    AuthlibKey,
    AuthlibSecretKey,
)


ALICE_KID = "did:example:alice#key-1"
BOB_KID = "did:example:bob#key-1"


@pytest.fixture
def alice_askar_key():
    yield Key.generate(KeyAlg.X25519)


@pytest.fixture
def bob_askar_key():
    yield Key.generate(KeyAlg.X25519)


@pytest.fixture
def alice_authlib_key(alice_askar_key: Key):
    yield OKPKey.import_key(json.loads(alice_askar_key.get_jwk_public()))


@pytest.fixture
def bob_authlib_key(bob_askar_key: Key):
    yield OKPKey.import_key(json.loads(bob_askar_key.get_jwk_secret()))


@pytest.fixture
def alice(alice_askar_key: Key, alice_authlib_key: OKPKey):
    yield (
        AskarSecretKey(alice_askar_key, ALICE_KID),
        AuthlibKey(alice_authlib_key, ALICE_KID),
    )


@pytest.fixture
def bob(bob_askar_key: Key, bob_authlib_key: OKPKey):
    yield AuthlibSecretKey(bob_authlib_key, BOB_KID), AskarKey(bob_askar_key, BOB_KID)


@pytest.fixture
def askar():
    yield AskarCryptoService()


@pytest.fixture
def authlib():
    yield AuthlibCryptoService()


@pytest.mark.asyncio
async def test_compat_ecdh_1pu(
    askar: AskarCryptoService,
    authlib: AuthlibCryptoService,
    alice: tuple[AskarSecretKey, AuthlibKey],
    bob: tuple[AuthlibSecretKey, AskarKey],
):
    """Test compabibility between Askar and Authlib.

    Alice uses Askar, Bob uses Authlib.
    """
    alice_sk, alice_pk = alice
    bob_sk, bob_pk = bob

    to_alice = b"Dear alice, please decrypt this"
    alice_enc_message = await authlib.ecdh_1pu_encrypt([alice_pk], bob_sk, to_alice)
    print(alice_enc_message)

    plaintext = await askar.ecdh_1pu_decrypt(alice_enc_message, alice_sk, bob_pk)
    assert plaintext == to_alice

    to_bob = b"Dear bob, please decrypt this"
    bob_enc_message = await askar.ecdh_1pu_encrypt([bob_pk], alice_sk, to_bob)

    print(bob_enc_message)

    plaintext = await authlib.ecdh_1pu_decrypt(bob_enc_message, bob_sk, alice_pk)
    assert plaintext == to_bob


@pytest.mark.asyncio
async def test_compat_ecdh_es(
    askar: AskarCryptoService,
    authlib: AuthlibCryptoService,
    alice: tuple[AskarSecretKey, AuthlibKey],
    bob: tuple[AuthlibSecretKey, AskarKey],
):
    """Test compabibility between Askar and Authlib.

    Alice uses Askar, Bob uses Authlib.
    """
    alice_sk, alice_pk = alice
    bob_sk, bob_pk = bob

    to_alice = b"Dear alice, please decrypt this"
    alice_enc_message = await authlib.ecdh_es_encrypt([alice_pk], to_alice)
    print(alice_enc_message)

    plaintext = await askar.ecdh_es_decrypt(alice_enc_message, alice_sk)
    assert plaintext == to_alice

    to_bob = b"Dear bob, please decrypt this"
    bob_enc_message = await askar.ecdh_es_encrypt([bob_pk], to_bob)

    print(bob_enc_message)

    plaintext = await authlib.ecdh_es_decrypt(bob_enc_message, bob_sk)
    assert plaintext == to_bob



"""Test PackagingService."""

import pytest

from aries_askar import Key, KeyAlg
from didcomm_messaging.crypto.backend.askar import AskarCryptoService, AskarSecretKey
from didcomm_messaging.crypto.backend.basic import InMemorySecretsManager
from didcomm_messaging.crypto.base import CryptoService
from didcomm_messaging.packaging import PackagingService
from didcomm_messaging.multiformats import multibase
from didcomm_messaging.multiformats import multicodec
from didcomm_messaging.resolver.peer import Peer2, Peer4
from didcomm_messaging.resolver import DIDResolver, PrefixResolver
from did_peer_2 import KeySpec, generate


@pytest.fixture
def secrets():
    """Fixture for secrets."""
    yield InMemorySecretsManager()


@pytest.fixture
def crypto():
    """Fixture for crypto."""
    yield AskarCryptoService()


@pytest.fixture
def resolver():
    yield PrefixResolver({"did:peer:2": Peer2(), "did:peer:4": Peer4()})


@pytest.fixture
def packaging():
    """Fixture for packaging."""
    yield PackagingService()


# TODO More thorough tests
@pytest.mark.asyncio
async def test_packer_basic(
    crypto: CryptoService,
    secrets: InMemorySecretsManager,
    resolver: DIDResolver,
    packaging: PackagingService,
):
    """Test basic packaging.

    This is a happy path test.
    """
    verkey = Key.generate(KeyAlg.ED25519)
    xkey = Key.generate(KeyAlg.X25519)
    did = generate(
        [
            KeySpec.verification(
                multibase.encode(
                    multicodec.wrap("ed25519-pub", verkey.get_public_bytes()),
                    "base58btc",
                )
            ),
            KeySpec.key_agreement(
                multibase.encode(
                    multicodec.wrap("x25519-pub", xkey.get_public_bytes()), "base58btc"
                )
            ),
        ],
        [],
    )
    await secrets.add_secret(AskarSecretKey(verkey, f"{did}#key-1"))
    await secrets.add_secret(AskarSecretKey(xkey, f"{did}#key-2"))
    message = b"hello world"
    packed = await packaging.pack(crypto, resolver, secrets, message, [did], did)
    unpacked, meta = await packaging.unpack(crypto, resolver, secrets, packed)
    assert unpacked == message



from aries_askar import Key, KeyAlg, Store
from pydid import VerificationMethod
import pytest

from aries_askar.bindings import generate_raw_key
from didcomm_messaging.crypto.backend.askar import (
    AskarCryptoService,
    AskarKey,
    AskarSecretKey,
    AskarSecretsManager,
)


ALICE_KID = "did:example:alice#key-1"
BOB_KID = "did:example:bob#key-1"
CAROL_KID = "did:example:carol#key-2"
MESSAGE = b"Expecto patronum"


@pytest.fixture
def crypto():
    yield AskarCryptoService()


@pytest.mark.asyncio
async def test_1pu_round_trip(crypto: AskarCryptoService):
    """Test ECDH-1PU (Authcrypt) round trip."""
    alg = KeyAlg.X25519
    alice_sk = Key.generate(alg)
    alice_pk = Key.from_jwk(alice_sk.get_jwk_public())
    bob_sk = Key.generate(alg)
    bob_pk = Key.from_jwk(bob_sk.get_jwk_public())
    bob_key = AskarKey(bob_sk, BOB_KID)
    bob_priv_key = AskarSecretKey(bob_sk, BOB_KID)
    alice_key = AskarKey(alice_sk, ALICE_KID)
    alice_priv_key = AskarSecretKey(alice_sk, ALICE_KID)

    enc_message = await crypto.ecdh_1pu_encrypt([bob_key], alice_priv_key, MESSAGE)

    plaintext = await crypto.ecdh_1pu_decrypt(enc_message, bob_priv_key, alice_key)
    assert plaintext == MESSAGE


@pytest.mark.asyncio
async def test_es_round_trip(crypto: AskarCryptoService):
    """Test ECDH-ES (Anoncrypt) round trip."""
    alg = KeyAlg.X25519
    alice_sk = Key.generate(alg)
    alice_pk = Key.from_jwk(alice_sk.get_jwk_public())
    bob_sk = Key.generate(alg)
    bob_pk = Key.from_jwk(bob_sk.get_jwk_public())
    bob_key = AskarKey(bob_sk, BOB_KID)
    bob_priv_key = AskarSecretKey(bob_sk, BOB_KID)
    alice_key = AskarKey(alice_sk, ALICE_KID)
    alice_priv_key = AskarSecretKey(alice_sk, ALICE_KID)

    enc_message = await crypto.ecdh_es_encrypt([bob_key], MESSAGE)

    plaintext = await crypto.ecdh_es_decrypt(enc_message, bob_priv_key)
    assert plaintext == MESSAGE


@pytest.mark.asyncio
async def test_askar_store():
    alg = KeyAlg.X25519
    alice_sk = Key.generate(alg)
    alice_key = AskarKey(alice_sk, ALICE_KID)

    raw_key = generate_raw_key()

    store = await Store.provision("sqlite://:memory:", "raw", raw_key, recreate=True)
    askar_secrets_manager = AskarSecretsManager(store)

    async with askar_secrets_manager.store.session() as session:
        await session.insert_key(alice_key.kid, alice_key.key)

        retrieved_key = await askar_secrets_manager.get_secret_by_kid(alice_key.kid)

        assert retrieved_key.kid == alice_key.kid  # type:ignore
        assert (
            retrieved_key.key.get_public_bytes()  # type:ignore
            == alice_key.key.get_public_bytes()
        )


@pytest.mark.parametrize(
    ["vm", "expected"],
    [
        [
            {
                "id": "#6LSqPZfn",
                "type": "X25519KeyAgreementKey2020",
                "publicKeyMultibase": "z6LSqPZfn9krvgXma2icTMKf2uVcYhKXsudCmPoUzqGYW24U",
                "controller": "did:example:1",
            },
            "did:example:1#6LSqPZfn",
        ],
        [
            {
                "id": "#6MkrCD1c",
                "type": "Ed25519VerificationKey2020",
                "publicKeyMultibase": "z6MkrCD1csqtgdj8sjrsu8jxcbeyP6m7LiK87NzhfWqio5yr",
                "controller": "did:example:2",
            },
            "did:example:2#6MkrCD1c",
        ],
        [
            {
                "id": "#1",
                "type": "Ed25519VerificationKey2018",
                "publicKeyBase58": "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG",
                "controller": "did:example:3",
            },
            "did:example:3#1",
        ],
        [
            {
                "id": "#1",
                "type": "Ed25519VerificationKey2018",
                "publicKeyBase58": "3dtu2WWtd5ELwRTJEPzmEJUYEp8Qq36N2QA24g9tFXK9",
                "controller": "did:peer:4zQmRMVzDUXhV64pfw3vFaDvyExjzW9oBXCF2n4zYCaHQFAT:zMx3zwMnDECV3GiFs8nmHr38TziMEEkcgFBEDH5PXQ8hxnMrwNfB9wTwskpJMjggeg8NF1jeDSK5772op2zLLdy8TGFCEiYQxpUvvku8qSCZx5Q8V9Li9mDp6WEqGabXLQ9GTinmyQHQyJ6TcfbHaTtJUFjHS962LFPdUwv3aDK673Pci2doTyHVTAsw4m5eToS2dKbtix9f7HNxwvixnbQucWNAWVAF6HTxFYRYmrRPDmeE8n7V1fXFkY7yvR6BWxKiWwHd8Vb1TbBBRStf5niRM2dUAyjJorTstPWSfG2pN5DsRF81NUd7Aif4EhNAQEJCTuAHxQ3rCnNkb9Pf7YTTxbt1t25YgDMioDi4uFhYcnTbHj7D74yNPC2Cfk6WasU69hMxj7Wxro58vtkA6hvDWGtnDyX4PzntBp3fn62R25HW2jadsZMiJpm5ufpYSktEFEHX6gGeF4KPgyU8b2hhyS3FKL4DULYLB6d6CZqUpwrJesGfDtFjfG1btbdmjd6Lm7FCbL3fU9E3AJWEmnFkg16vARiQ1CrzeS9SyNtybKCk4",
            },
            "did:peer:4zQmRMVzDUXhV64pfw3vFaDvyExjzW9oBXCF2n4zYCaHQFAT:zMx3zwMnDECV3GiFs8nmHr38TziMEEkcgFBEDH5PXQ8hxnMrwNfB9wTwskpJMjggeg8NF1jeDSK5772op2zLLdy8TGFCEiYQxpUvvku8qSCZx5Q8V9Li9mDp6WEqGabXLQ9GTinmyQHQyJ6TcfbHaTtJUFjHS962LFPdUwv3aDK673Pci2doTyHVTAsw4m5eToS2dKbtix9f7HNxwvixnbQucWNAWVAF6HTxFYRYmrRPDmeE8n7V1fXFkY7yvR6BWxKiWwHd8Vb1TbBBRStf5niRM2dUAyjJorTstPWSfG2pN5DsRF81NUd7Aif4EhNAQEJCTuAHxQ3rCnNkb9Pf7YTTxbt1t25YgDMioDi4uFhYcnTbHj7D74yNPC2Cfk6WasU69hMxj7Wxro58vtkA6hvDWGtnDyX4PzntBp3fn62R25HW2jadsZMiJpm5ufpYSktEFEHX6gGeF4KPgyU8b2hhyS3FKL4DULYLB6d6CZqUpwrJesGfDtFjfG1btbdmjd6Lm7FCbL3fU9E3AJWEmnFkg16vARiQ1CrzeS9SyNtybKCk4#1",
        ],
        [
            {
                "id": "#z6MkrmNwty5ajKtFqc1U48oL2MMLjWjartwc5sf2AihZwXDN",
                "type": "Ed25519VerificationKey2018",
                "publicKeyBase58": "DK7uJiq9PnPnj7AmNZqVBFoLuwTjT1hFPrk6LSjZ2JRz",
                "controller": "did:example:4",
            },
            "did:example:4#z6MkrmNwty5ajKtFqc1U48oL2MMLjWjartwc5sf2AihZwXDN",
        ],
    ],
)
@pytest.mark.asyncio
async def test_key_from_verification_method(vm, expected):
    vm = VerificationMethod.deserialize(vm)

    key = AskarCryptoService.verification_method_to_public_key(vm)
    print(key.kid)
    assert type(key) == AskarKey
    assert key.kid == expected


@pytest.mark.xfail
@pytest.mark.asyncio
async def test_key_from_verification_method_unsupported():
    # Fails to deserialize, don't know what we can do about that (or if it's even valuable to have this test here)
    vm = VerificationMethod.deserialize(
        {
            # Here's a case that we probably don't support right now:
            "type": "Ed25519VerificationKey2018",
            "publicKeyJwk": {
                "kty": "OKP",
                "crv": "Ed25519",
                "x": "UTBElpNSZB8dS_R9rzWnWB-ozdtL7Sz96RQZhwnzur8",
            },
            "id": "#z6MkjvBkt8ETnxXGBFPSGgYKb43q7oNHLX8BiYSPcXVG6gY6",
        }
    )

    key = AskarCryptoService.verification_method_to_public_key(vm)
    print(key.kid)
    assert type(key) == AskarKey
    # assert key.kid == expected



"""Askar backend for DIDComm Messaging."""

from collections import OrderedDict
import hashlib
import json
from typing import Optional, Sequence, Union

from pydid import VerificationMethod

from didcomm_messaging.crypto.base import (
    CryptoService,
    CryptoServiceError,
    PublicKey,
    SecretKey,
    SecretsManager,
)
from didcomm_messaging.crypto.jwe import JweBuilder, JweEnvelope, JweRecipient, b64url
from didcomm_messaging.multiformats import multibase, multicodec

try:
    from aries_askar import Key, ecdh, AskarError, KeyAlg, Store
except ImportError:
    raise ImportError("Askar backend requires the 'askar' extra to be installed")


class AskarKey(PublicKey):
    """Public key implementation for Askar."""

    codec_to_alg = {
        "ed25519-pub": KeyAlg.ED25519,
        "x25519-pub": KeyAlg.X25519,
        "secp256k1-pub": KeyAlg.K256,
    }
    alg_to_codec = {v: k for k, v in codec_to_alg.items()}

    type_to_alg = {
        "Ed25519VerificationKey2018": KeyAlg.ED25519,
        "X25519KeyAgreementKey2019": KeyAlg.X25519,
        "Ed25519VerificationKey2020": KeyAlg.ED25519,
        "X25519KeyAgreementKey2020": KeyAlg.X25519,
        "EcdsaSecp256k1VerificationKey2019": KeyAlg.K256,
    }

    def __init__(self, key: Key, kid: str):
        """Initialize a new AskarKey instance."""
        self.key = key
        self._kid = kid
        self._multikey = self.key_to_multikey(key)

    @classmethod
    def key_to_multikey(cls, key: Key) -> str:
        """Get a multikey from an Askar Key instance."""
        codec = cls.alg_to_codec.get(key.algorithm)
        if not codec:
            raise ValueError("Unsupported key type")

        return multibase.encode(
            multicodec.wrap(multicodec.multicodec(codec), key.get_public_bytes()),
            "base58btc",
        )

    @classmethod
    def multikey_to_key(cls, multikey: str) -> Key:
        """Convert a multibase-encoded key to an Askar Key instance."""
        decoded = multibase.decode(multikey)
        codec, key = multicodec.unwrap(decoded)
        alg = cls.codec_to_alg.get(codec.name)
        if not alg:
            raise ValueError("Unsupported key type: {codec.name}")
        try:
            return Key.from_public_bytes(alg, key)
        except AskarError as err:
            raise ValueError("Invalid key") from err

    @classmethod
    def from_verification_method(cls, vm: VerificationMethod) -> "AskarKey":
        """Create a Key instance from a DID Document Verification Method."""
        if not vm.id.did:
            kid = vm.id.as_absolute(vm.controller)
        else:
            kid = vm.id

        if vm.type == "Multikey":
            multikey = vm.public_key_multibase
            if not multikey:
                raise ValueError("Multikey verification method missing key")

            key = cls.multikey_to_key(multikey)
            return cls(key, kid)

        if vm.type == "JsonWebKey2020":
            jwk = vm.public_key_jwk
            if not jwk:
                raise ValueError("JWK verification method missing key")
            try:
                key = Key.from_jwk(jwk)
            except AskarError as err:
                raise ValueError("Invalid JWK") from err
            return cls(key, kid)

        alg = cls.type_to_alg.get(vm.type)
        if not alg:
            raise ValueError("Unsupported verification method type: {vm_type}")

        key_bytes = cls.key_bytes_from_verification_method(vm)
        key = Key.from_public_bytes(alg, key_bytes)
        return cls(key, kid)

    @property
    def kid(self) -> str:
        """Get the key ID."""
        return self._kid

    @property
    def multikey(self) -> str:
        """Get the key in multibase format."""
        return self._multikey


class AskarSecretKey(SecretKey):
    """Secret key implementation for Askar."""

    def __init__(self, key: Key, kid: str):
        """Initialize a new AskarSecretKey instance."""
        self.key = key
        self._kid = kid

    @property
    def kid(self) -> str:
        """Get the key ID."""
        return self._kid


class AskarCryptoService(CryptoService[AskarKey, AskarSecretKey]):
    """CryptoService backend implemented using Askar."""

    async def ecdh_es_encrypt(self, to_keys: Sequence[AskarKey], message: bytes) -> bytes:
        """Encode a message into DIDComm v2 anonymous encryption."""
        builder = JweBuilder(with_flatten_recipients=False)

        alg_id = "ECDH-ES+A256KW"
        enc_id = "XC20P"
        enc_alg = KeyAlg.XC20P
        wrap_alg = KeyAlg.A256KW

        if not to_keys:
            raise ValueError("No message recipients")

        try:
            cek = Key.generate(enc_alg)
        except AskarError:
            raise CryptoServiceError("Error creating content encryption key")

        apv = []
        for recip_key in to_keys:
            apv.append(recip_key.kid)
        apv.sort()
        apv = hashlib.sha256((".".join(apv)).encode()).digest()

        for recip_key in to_keys:
            try:
                epk = Key.generate(recip_key.key.algorithm, ephemeral=True)
            except AskarError:
                raise CryptoServiceError("Error creating ephemeral key")
            enc_key = ecdh.EcdhEs(alg_id, None, apv).sender_wrap_key(  # type: ignore
                wrap_alg, epk, recip_key.key, cek
            )
            builder.add_recipient(
                JweRecipient(
                    encrypted_key=enc_key.ciphertext,
                    header={
                        "kid": recip_key.kid,
                        "epk": json.loads(epk.get_jwk_public()),
                    },
                )
            )

        builder.set_protected(
            OrderedDict(
                [
                    ("typ", "application/didcomm-encrypted+json"),
                    ("alg", alg_id),
                    ("enc", enc_id),
                    ("apv", b64url(apv)),
                ]
            )
        )
        try:
            payload = cek.aead_encrypt(message, aad=builder.protected_bytes)
        except AskarError:
            raise CryptoServiceError("Error encrypting message payload")
        builder.set_payload(payload.ciphertext, payload.nonce, payload.tag)

        return builder.build().to_json().encode("utf-8")

    async def ecdh_es_decrypt(
        self,
        enc_message: Union[str, bytes],
        recip_key: AskarSecretKey,
    ) -> bytes:
        """Decode a message from DIDComm v2 anonymous encryption."""
        if isinstance(enc_message, bytes):
            wrapper = enc_message.decode("utf-8")
        wrapper = JweEnvelope.from_json(enc_message)

        alg_id = wrapper.protected.get("alg")

        if alg_id and alg_id in ("ECDH-ES+A128KW", "ECDH-ES+A256KW"):
            wrap_alg = alg_id[8:]
        else:
            raise CryptoServiceError(
                f"Missing or unsupported ECDH-ES algorithm: {alg_id}"
            )

        recip = wrapper.get_recipient(recip_key.kid)
        if not recip:
            raise CryptoServiceError(f"Recipient header not found: {recip_key.kid}")

        enc_alg = recip.header.get("enc")
        if not enc_alg or enc_alg not in (
            "A128GCM",
            "A256GCM",
            "A128CBC-HS256",
            "A256CBC-HS512",
            "XC20P",
        ):
            raise CryptoServiceError(f"Unsupported ECDH-ES content encryption: {enc_alg}")

        epk_header = recip.header.get("epk")
        if not epk_header:
            raise CryptoServiceError("Missing ephemeral key")

        try:
            epk = Key.from_jwk(epk_header)
        except AskarError:
            raise CryptoServiceError("Error loading ephemeral key")

        try:
            cek = ecdh.EcdhEs(alg_id, None, wrapper.apv_bytes).receiver_unwrap_key(  # type: ignore
                wrap_alg,
                enc_alg,
                epk,
                recip_key.key,
                recip.encrypted_key,
            )
        except AskarError:
            raise CryptoServiceError("Error decrypting content encryption key")

        try:
            plaintext = cek.aead_decrypt(
                wrapper.ciphertext,
                nonce=wrapper.iv,
                tag=wrapper.tag,
                aad=wrapper.combined_aad,
            )
        except AskarError:
            raise CryptoServiceError("Error decrypting message payload")

        return plaintext

    async def ecdh_1pu_encrypt(
        self,
        to_keys: Sequence[AskarKey],
        sender_key: AskarSecretKey,
        message: bytes,
    ) -> bytes:
        """Encode a message into DIDComm v2 authenticated encryption."""
        builder = JweBuilder(with_flatten_recipients=False)

        alg_id = "ECDH-1PU+A256KW"
        enc_id = "A256CBC-HS512"
        enc_alg = KeyAlg.A256CBC_HS512
        wrap_alg = KeyAlg.A256KW
        agree_alg = sender_key.key.algorithm

        if not to_keys:
            raise CryptoServiceError("No message recipients")

        try:
            cek = Key.generate(enc_alg)
        except AskarError:
            raise CryptoServiceError("Error creating content encryption key")

        try:
            epk = Key.generate(agree_alg, ephemeral=True)
        except AskarError:
            raise CryptoServiceError("Error creating ephemeral key")

        apu = sender_key.kid
        apv = []
        for recip_key in to_keys:
            if agree_alg:
                if agree_alg != recip_key.key.algorithm:
                    raise CryptoServiceError("Recipient key types must be consistent")
            else:
                agree_alg = recip_key.key.algorithm
            apv.append(recip_key.kid)
        apv.sort()
        apv = hashlib.sha256((".".join(apv)).encode()).digest()

        builder.set_protected(
            OrderedDict(
                [
                    ("typ", "application/didcomm+encrypted"),
                    ("alg", alg_id),
                    ("enc", enc_id),
                    ("apu", b64url(apu)),
                    ("apv", b64url(apv)),
                    ("epk", json.loads(epk.get_jwk_public())),
                    ("skid", sender_key.kid),
                ]
            )
        )
        try:
            payload = cek.aead_encrypt(message, aad=builder.protected_bytes)
        except AskarError:
            raise CryptoServiceError("Error encrypting message payload")
        builder.set_payload(payload.ciphertext, payload.nonce, payload.tag)

        for recip_key in to_keys:
            enc_key = ecdh.Ecdh1PU(alg_id, apu, apv).sender_wrap_key(
                wrap_alg, epk, sender_key.key, recip_key.key, cek, cc_tag=payload.tag
            )
            builder.add_recipient(
                JweRecipient(
                    encrypted_key=enc_key.ciphertext, header={"kid": recip_key.kid}
                )
            )

        return builder.build().to_json().encode("utf-8")

    async def ecdh_1pu_decrypt(
        self,
        enc_message: Union[str, bytes],
        recip_key: AskarSecretKey,
        sender_key: AskarKey,
    ):
        """Decode a message from DIDComm v2 authenticated encryption."""
        if isinstance(enc_message, bytes):
            wrapper = enc_message.decode("utf-8")
        wrapper = JweEnvelope.from_json(enc_message)

        alg_id = wrapper.protected.get("alg")
        if alg_id and alg_id in ("ECDH-1PU+A128KW", "ECDH-1PU+A256KW"):
            wrap_alg = alg_id[9:]
        else:
            raise CryptoServiceError(f"Unsupported ECDH-1PU algorithm: {alg_id}")

        enc_alg = wrapper.protected.get("enc")
        if not enc_alg or enc_alg not in ("A128CBC-HS256", "A256CBC-HS512"):
            raise CryptoServiceError(
                f"Unsupported ECDH-1PU content encryption: {enc_alg}"
            )

        recip = wrapper.get_recipient(recip_key.kid)
        if not recip:
            raise CryptoServiceError(f"Recipient header not found: {recip_key.kid}")

        epk_header = recip.header.get("epk")
        if not epk_header:
            raise CryptoServiceError("Missing ephemeral key")

        try:
            epk = Key.from_jwk(epk_header)
        except AskarError:
            raise CryptoServiceError("Error loading ephemeral key")

        try:
            cek = ecdh.Ecdh1PU(
                alg_id, wrapper.apu_bytes, wrapper.apv_bytes
            ).receiver_unwrap_key(  # type: ignore
                wrap_alg,
                enc_alg,
                epk,
                sender_key.key,
                recip_key.key,
                recip.encrypted_key,
                cc_tag=wrapper.tag,
            )
        except AskarError as err:
            raise CryptoServiceError("Error decrypting content encryption key") from err

        try:
            plaintext = cek.aead_decrypt(
                wrapper.ciphertext,
                nonce=wrapper.iv,
                tag=wrapper.tag,
                aad=wrapper.combined_aad,
            )
        except AskarError:
            raise CryptoServiceError("Error decrypting message payload")

        return plaintext

    @classmethod
    def verification_method_to_public_key(cls, vm: VerificationMethod) -> AskarKey:
        """Convert a verification method into a public key."""
        return AskarKey.from_verification_method(vm)


class AskarSecretsManager(SecretsManager[AskarSecretKey]):
    """Askar KMS with an Askar Store for secrets management."""

    def __init__(self, store: Store):
        """Initialize a new Askar instance."""
        self.store = store

    async def get_secret_by_kid(self, kid: str) -> Optional[AskarSecretKey]:
        """Fetch a public key by key ID."""
        async with self.store.session() as session:
            key_entry = await session.fetch_key(kid)
            if not key_entry:
                return None

        # cached_property doesn't play nice with pyright
        return AskarKey(key_entry.key, kid)  # type: ignore



"""DIDComm Messaging."""

from dataclasses import dataclass
import json
from typing import Generic, Optional, List

from pydid.service import DIDCommV2Service

from didcomm_messaging.crypto import CryptoService, SecretsManager, P, S
from didcomm_messaging.packaging import PackagingService
from didcomm_messaging.resolver import DIDResolver
from didcomm_messaging.routing import RoutingService


@dataclass
class PackResult:
    """Result of packing a message."""

    message: bytes
    target_services: List[DIDCommV2Service]

    def get_endpoint(self, protocol: str) -> str:
        """Get the first matching endpoint to send the message to."""
        return self.get_service(protocol).service_endpoint.uri

    def get_service(self, protocol: str) -> DIDCommV2Service:
        """Get the first matching service to send the message to."""
        return self.filter_services_by_protocol(protocol)[0]

    def filter_services_by_protocol(self, protocol: str) -> List[DIDCommV2Service]:
        """Get all services that start with a specific uri protocol."""
        return [
            service
            for service in self.target_services
            if service.service_endpoint.uri.startswith(protocol)
        ]


@dataclass
class UnpackResult:
    """Result of unpacking a message."""

    message: dict
    encrytped: bool
    authenticated: bool
    recipient_kid: str
    sender_kid: Optional[str] = None


class DIDCommMessagingService(Generic[P, S]):
    """Main entrypoint for DIDComm Messaging."""

    def service_to_target(self, service: DIDCommV2Service) -> str:
        """Convert a service to a target uri.

        This is a very simple implementation that just returns the first one.
        """
        if isinstance(service.service_endpoint, list):
            service_endpoint = service.service_endpoint[0]
        else:
            service_endpoint = service.service_endpoint

        return service_endpoint.uri

    async def pack(
        self,
        crypto: CryptoService[P, S],
        resolver: DIDResolver,
        secrets: SecretsManager[S],
        packaging: PackagingService[P, S],
        routing: RoutingService,
        message: dict,
        to: str,
        frm: Optional[str] = None,
        **options,
    ):
        """Pack a message."""
        # TODO crypto layer permits packing to multiple recipients; should we as well?

        encoded_message = await packaging.pack(
            crypto,
            resolver,
            secrets,
            json.dumps(message).encode(),
            [to],
            frm,
            **options,
        )

        forward, services = await routing.prepare_forward(
            crypto, packaging, resolver, secrets, to, encoded_message
        )
        return PackResult(forward, services)

    async def unpack(
        self,
        crypto: CryptoService[P, S],
        resolver: DIDResolver,
        secrets: SecretsManager[S],
        packaging: PackagingService[P, S],
        encoded_message: bytes,
        **options,
    ) -> UnpackResult:
        """Unpack a message."""
        unpacked, metadata = await packaging.unpack(
            crypto, resolver, secrets, encoded_message, **options
        )
        message = json.loads(unpacked.decode())
        return UnpackResult(
            message,
            encrytped=bool(metadata.method),
            authenticated=bool(metadata.sender_kid),
            recipient_kid=metadata.recip_key.kid,
            sender_kid=metadata.sender_kid,
        )


class DIDCommMessaging(Generic[P, S]):
    """Main entrypoint for DIDComm Messaging."""

    def __init__(
        self,
        crypto: CryptoService[P, S],
        secrets: SecretsManager[S],
        resolver: DIDResolver,
        packaging: PackagingService[P, S],
        routing: RoutingService,
    ):
        """Initialize the DIDComm Messaging service."""
        self.crypto = crypto
        self.secrets = secrets
        self.resolver = resolver
        self.packaging = packaging
        self.routing = routing
        self.dmp = DIDCommMessagingService()

    async def pack(
        self,
        message: dict,
        to: str,
        frm: Optional[str] = None,
        **options,
    ) -> PackResult:
        """Pack a message."""
        return await self.dmp.pack(
            self.crypto,
            self.resolver,
            self.secrets,
            self.packaging,
            self.routing,
            message,
            to,
            frm,
            **options,
        )

    async def unpack(
        self,
        encoded_message: bytes,
        **options,
    ) -> UnpackResult:
        """Unpack a message."""
        return await self.dmp.unpack(
            self.crypto,
            self.resolver,
            self.secrets,
            self.packaging,
            encoded_message,
            **options,
        )



"""did:peer resolver."""

from didcomm_messaging.resolver import DIDResolver

try:
    from did_peer_2 import resolve as resolve_peer_2
    from did_peer_2 import PATTERN as peer_2_pattern
    from did_peer_4 import resolve as resolve_peer_4
    from did_peer_4 import LONG_PATTERN as peer_4_pattern_long
    from did_peer_4 import SHORT_PATTERN as peer_4_pattern_short
except ImportError:
    raise ImportError(
        "did-peer-2 and did-peer-4 are required for did:peer resolution; "
        "install the did_peer extra"
    )


class Peer2(DIDResolver):
    """did:peer:2 resolver."""

    async def is_resolvable(self, did: str) -> bool:
        """Check to see if a DID is resolvable."""
        return peer_2_pattern.match(did)

    async def resolve(self, did: str) -> dict:
        """Resolve a did:peer:2 DID."""
        return resolve_peer_2(did)


class Peer4(DIDResolver):
    """did:peer:4 resolver."""

    async def is_resolvable(self, did: str) -> bool:
        """Check to see if a DID is resolvable."""
        return peer_4_pattern_short.match(did) or peer_4_pattern_long.match(did)

    async def resolve(self, did: str) -> dict:
        """Resolve a did:peer:4 DID."""
        return resolve_peer_4(did)



"""JSON Web Encryption utilities."""

import binascii
from dataclasses import dataclass, field
import json
from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

from didcomm_messaging.multiformats.multibase import Base64UrlEncoder

IDENT_ENC_KEY = "encrypted_key"
IDENT_HEADER = "header"
IDENT_PROTECTED = "protected"
IDENT_RECIPIENTS = "recipients"


_base64url = Base64UrlEncoder()


def b64url(value: Union[bytes, str]) -> str:
    """Encode a string or bytes value as unpadded base64-URL."""
    if isinstance(value, str):
        value = value.encode("utf-8")
    return _base64url.encode(value)


def from_b64url(value: str) -> bytes:
    """Decode an unpadded base64-URL value."""
    try:
        return _base64url.decode(value)
    except binascii.Error:
        raise ValueError("Error decoding base64 value")


class JweRecipient:
    """A single message recipient."""

    def __init__(self, *, encrypted_key: bytes, header: Optional[dict] = None):
        """Initialize the JWE recipient."""
        self.encrypted_key = encrypted_key
        self.header = header or {}

    @classmethod
    def deserialize(cls, entry: Mapping[str, Any]) -> "JweRecipient":
        """Deserialize a JWE recipient from a mapping."""
        if "encrypted_key" not in entry:
            raise ValueError("Invalid JWE recipient: missing encrypted_key")

        encrypyted_key = from_b64url(entry["encrypted_key"])

        if "header" in entry:
            if not isinstance(entry["header"], dict):
                raise ValueError("Invalid JWE recipient: invalid header")
            return cls(encrypted_key=encrypyted_key, header=entry["header"])

        return cls(encrypted_key=encrypyted_key)

    def serialize(self) -> dict:
        """Serialize the JWE recipient to a mapping."""
        ret: OrderedDict[str, Any] = OrderedDict(
            [("encrypted_key", b64url(self.encrypted_key))]
        )
        if self.header:
            ret["header"] = self.header
        return ret


class JweBuilder:
    """Builder for JWE envelopes."""

    def __init__(
        self,
        with_protected_recipients: bool = False,
        with_flatten_recipients: bool = True,
    ):
        """Initialize the JWE builder."""
        self._with_protected_recipients = with_protected_recipients
        self._with_flatten_recipients = with_flatten_recipients
        self._recipients: List[JweRecipient] = []
        self._protected: Optional[OrderedDict] = None
        self._unprotected: Optional[OrderedDict] = None
        self._protected_b64: Optional[bytes] = None
        self._ciphertext: Optional[bytes] = None
        self._iv: Optional[bytes] = None
        self._tag: Optional[bytes] = None
        self._aad: Optional[bytes] = None

    def add_recipient(self, recip: JweRecipient):
        """Add a recipient to the JWE envelope."""
        self._recipients.append(recip)

    @property
    def recipients_json(self) -> List[Dict[str, Any]]:
        """Encode the current recipients for JSON."""
        return [recip.serialize() for recip in self._recipients]

    @property
    def protected_bytes(self) -> bytes:
        """Access the protected data encoded as bytes."""
        assert self._protected_b64
        return self._protected_b64

    def set_protected(
        self,
        protected: Mapping[str, Any],
    ):
        """Set the protected headers of the JWE envelope."""
        protected = OrderedDict(protected.items())
        if self._with_protected_recipients:
            recipients = self.recipients_json
            if self._with_flatten_recipients and len(recipients) == 1:
                protected.update(recipients[0])
            elif recipients:
                protected[IDENT_RECIPIENTS] = recipients
            else:
                raise ValueError("Missing message recipients")
        self._protected = protected
        self._protected_b64 = b64url(json.dumps(protected)).encode("utf-8")

    def set_unprotected(self, unprotected: OrderedDict):
        """Set the unprotected headers of the JWE envelope."""
        self._unprotected = unprotected

    def set_payload(
        self, ciphertext: bytes, iv: bytes, tag: bytes, aad: Optional[bytes] = None
    ):
        """Set the payload of the JWE envelope."""
        self._ciphertext = ciphertext
        self._iv = iv
        self._tag = tag
        self._aad = aad

    def build(self):
        """Build the JWE envelope."""
        if not self._ciphertext:
            raise ValueError("Missing payload for JWE")

        if not self._iv:
            raise ValueError("Missing iv (nonce) for JWE")

        if not self._tag:
            raise ValueError("Missing tag for JWE")

        if not self._recipients:
            raise ValueError("Missing recipients for JWE")

        if not self._protected:
            raise ValueError("Missing protected headers for JWE")

        if not self._protected_b64:
            raise ValueError("Missing encoded protected headers for JWE")

        return JweEnvelope(
            protected=self._protected,
            protected_b64=self._protected_b64,
            ciphertext=self._ciphertext,
            iv=self._iv,
            tag=self._tag,
            aad=self._aad,
            unprotected=self._unprotected,
            recipients=self._recipients,
            with_flatten_recipients=self._with_flatten_recipients,
            with_protected_recipients=self._with_protected_recipients,
        )


@dataclass
class JweEnvelope:
    """JWE envelope instance."""

    recipients: List[JweRecipient]
    protected: dict
    protected_b64: bytes
    ciphertext: bytes
    iv: bytes
    tag: bytes
    aad: Optional[bytes] = None
    unprotected: Optional[dict] = field(default_factory=OrderedDict)
    with_protected_recipients: bool = False
    with_flatten_recipients: bool = True

    @classmethod
    def from_json(cls, message: Union[bytes, str]) -> "JweEnvelope":
        """Decode a JWE envelope from a JSON string or bytes value."""
        try:
            return cls._deserialize(json.loads(message))
        except json.JSONDecodeError:
            raise ValueError("Invalid JWE: not JSON")

    @classmethod
    def deserialize(cls, message: Mapping[str, Any]) -> "JweEnvelope":  # noqa: C901
        """Deserialize a JWE envelope from a mapping."""
        # Basic validation

        if not isinstance(message, dict):
            raise ValueError("Invalid JWE: not a mapping")

        # Validate protected
        if "protected" not in message:
            raise ValueError("Invalid JWE: missing protected header")

        if not isinstance(message["protected"], str):
            raise ValueError("Invalid JWE: invalid protected header")

        # validate unprotected
        if "unprotected" in message and not isinstance(message["unprotected"], dict):
            raise ValueError("Invalid JWE: invalid unprotected header")

        # validate recipients
        if "recipients" in message:
            if not isinstance(message["recipients"], list):
                raise ValueError("Invalid JWE: invalid recipients")
            for recip in message["recipients"]:
                if not isinstance(recip, dict):
                    raise ValueError("Invalid JWE: invalid recipient")

        # validate ciphertext
        if "ciphertext" not in message:
            raise ValueError("Invalid JWE: missing ciphertext")

        if not isinstance(message["ciphertext"], str):
            raise ValueError("Invalid JWE: invalid ciphertext")

        # validate iv
        if "iv" not in message:
            raise ValueError("Invalid JWE: missing iv")

        if not isinstance(message["iv"], str):
            raise ValueError("Invalid JWE: invalid iv")

        # validate tag
        if "tag" not in message:
            raise ValueError("Invalid JWE: missing tag")

        if not isinstance(message["tag"], str):
            raise ValueError("Invalid JWE: invalid tag")

        # validate aad
        if "aad" in message and not isinstance(message["aad"], str):
            raise ValueError("Invalid JWE: invalid aad")

        # validate header
        if "header" in message:
            if not isinstance(message["header"], dict):
                raise ValueError("Invalid JWE: invalid header")

        # validate encrypted_key
        if "encrypted_key" in message and not isinstance(message["encrypted_key"], str):
            raise ValueError("Invalid JWE: invalid encrypted_key")

        return cls._deserialize(message)

    @classmethod
    def _deserialize(cls, parsed: Mapping[str, Any]) -> "JweEnvelope":  # noqa: C901
        protected_b64 = parsed[IDENT_PROTECTED]
        try:
            protected: dict = json.loads(from_b64url(protected_b64))
        except json.JSONDecodeError:
            raise ValueError("Invalid JWE: invalid JSON for protected headers") from None
        unprotected = parsed.get("unprotected") or {}
        if protected.keys() & unprotected.keys():
            raise ValueError("Invalid JWE: duplicate header")

        encrypted_key = protected.get(IDENT_ENC_KEY) or parsed.get(IDENT_ENC_KEY)
        recipients = None
        protected_recipients = False
        flat_recipients = False

        if IDENT_RECIPIENTS in protected:
            recipients = protected.pop(IDENT_RECIPIENTS)
            if IDENT_RECIPIENTS in parsed:
                raise ValueError("Invalid JWE: duplicate recipients block")
            protected_recipients = True
        elif IDENT_RECIPIENTS in parsed:
            recipients = parsed[IDENT_RECIPIENTS]

        if IDENT_ENC_KEY in protected:
            encrypted_key = from_b64url(protected.pop(IDENT_ENC_KEY))
            header = protected.pop(IDENT_HEADER) if IDENT_HEADER in protected else None
            protected_recipients = True
        elif IDENT_ENC_KEY in parsed:
            encrypted_key = parsed[IDENT_ENC_KEY]
            header = parsed.get(IDENT_HEADER)
        else:
            header = None

        if recipients:
            if encrypted_key:
                raise ValueError("Invalid JWE: flattened form with 'recipients'")
            recipients = [JweRecipient.deserialize(recip) for recip in recipients]
        elif encrypted_key:
            recipients = [
                JweRecipient(
                    encrypted_key=encrypted_key,
                    header=header,
                )
            ]
            flat_recipients = True
        else:
            raise ValueError("Invalid JWE: no recipients")

        all_h = protected.keys() | unprotected.keys()
        for recip in recipients:
            if recip.header and recip.header.keys() & all_h:
                raise ValueError("Invalid JWE: duplicate header")

        ciphertext = from_b64url(parsed["ciphertext"])
        iv = from_b64url(parsed["iv"])
        tag = from_b64url(parsed["tag"])
        aad = from_b64url(parsed["aad"]) if "aad" in parsed else None

        inst = cls(
            recipients=recipients,
            protected=protected,
            protected_b64=protected_b64,
            unprotected=unprotected,
            ciphertext=ciphertext,
            iv=iv,
            tag=tag,
            aad=aad,
            with_protected_recipients=protected_recipients,
            with_flatten_recipients=flat_recipients,
        )

        return inst

    def serialize(self) -> dict:  # noqa: C901
        """Serialize the JWE envelope to a mapping."""
        if self.protected_b64 is None:
            raise ValueError("Missing protected: use set_protected")
        if self.ciphertext is None:
            raise ValueError("Missing ciphertext for JWE")
        if self.iv is None:
            raise ValueError("Missing iv (nonce) for JWE")
        if self.tag is None:
            raise ValueError("Missing tag for JWE")
        env = OrderedDict()
        env["protected"] = self.protected_b64.decode("utf-8")
        if self.unprotected:
            env["unprotected"] = self.unprotected.copy()
        if not self.with_protected_recipients:
            recipients = self.recipients_json
            if self.with_flatten_recipients and len(recipients) == 1:
                for k in recipients[0]:
                    env[k] = recipients[0][k]
            elif recipients:
                env[IDENT_RECIPIENTS] = recipients
            else:
                raise ValueError("Missing message recipients")
        env["iv"] = b64url(self.iv)
        env["ciphertext"] = b64url(self.ciphertext)
        env["tag"] = b64url(self.tag)
        if self.aad:
            env["aad"] = b64url(self.aad)
        return env

    def to_json(self) -> str:
        """Serialize the JWE envelope to a JSON string."""
        return json.dumps(self.serialize())

    def get_recipients(self) -> Iterable[JweRecipient]:
        """Accessor for an iterator over the JWE recipients.

        The headers for each recipient include protected and unprotected headers from the
        outer envelope.
        """
        header = self.protected.copy()
        header.update(self.unprotected or {})
        for recip in self.recipients:
            if recip.header:
                recip_h = header.copy()
                recip_h.update(recip.header)
                yield JweRecipient(encrypted_key=recip.encrypted_key, header=recip_h)
            else:
                yield JweRecipient(encrypted_key=recip.encrypted_key, header=header)

    @property
    def recipients_json(self) -> List[Dict[str, Any]]:
        """Encode the current recipients for JSON."""
        return [recip.serialize() for recip in self.recipients]

    @property
    def recipient_key_ids(self) -> Iterable[str]:
        """Accessor for an iterator over the JWE recipient key identifiers."""
        for recip in self.recipients:
            if recip.header and "kid" in recip.header:
                yield recip.header["kid"]

    def get_recipient(self, kid: str) -> JweRecipient:
        """Find a recipient by key ID."""
        for recip in self.recipients:
            if recip.header and recip.header.get("kid") == kid:
                header = self.protected.copy()
                header.update(self.unprotected or {})
                header.update(recip.header)
                return JweRecipient(encrypted_key=recip.encrypted_key, header=header)
        raise ValueError(f"Unknown recipient: {kid}")

    @property
    def combined_aad(self) -> bytes:
        """Accessor for the additional authenticated data."""
        aad = self.protected_b64
        if self.aad:
            aad += b"." + b64url(self.aad).encode("utf-8")
        return aad

    @property
    def apu_bytes(self) -> bytes:
        """Accessor for the Agreement PartyUInfo."""
        if "apu" in self.protected:
            return from_b64url(self.protected["apu"])
        raise ValueError("Missing apu")

    @property
    def apv_bytes(self) -> bytes:
        """Accessor for the Agreement PartyVInfo."""
        if "apv" in self.protected:
            return from_b64url(self.protected["apv"])
        raise ValueError("Missing apv")



"""Basic Crypto Implementations."""

from typing import Optional
from didcomm_messaging.crypto.base import S, SecretsManager


class InMemorySecretsManager(SecretsManager[S]):
    """In Memory Secrets Manager."""

    def __init__(self, secrets: Optional[dict] = None):
        """Initialize the InMemorySecretsManager."""
        self.secrets = secrets or {}

    async def get_secret_by_kid(self, kid: str) -> Optional[S]:
        """Get a secret by its kid."""
        return self.secrets.get(kid)

    async def add_secret(self, secret: S) -> None:
        """Add a secret to the secrets manager."""
        self.secrets[secret.kid] = secret



"""RoutingService interface."""

import json
import uuid

from typing import Tuple, List, Dict, Any
from pydid.service import DIDCommV2Service
from didcomm_messaging.crypto.base import P, S, CryptoService, SecretsManager
from didcomm_messaging.packaging import PackagingService
from didcomm_messaging.resolver import DIDResolver


class RoutingServiceError(Exception):
    """Raised when an error occurs in the RoutingService."""


class RoutingService:
    """RoutingService."""

    async def _resolve_services(
        self, resolver: DIDResolver, to: str
    ) -> List[DIDCommV2Service]:
        if not await resolver.is_resolvable(to):
            return []
        did_doc = await resolver.resolve_and_parse(to)
        services = []
        if did_doc.service:  # service is not guaranteed to exist
            for did_service in did_doc.service:
                if did_service.type != "DIDCommMessaging":
                    continue
                if "didcomm/v2" in did_service.service_endpoint.accept:
                    services.append(did_service)
        if not services:
            return []
        return services

    async def is_forwardable_service(
        self, resolver: DIDResolver, service: DIDCommV2Service
    ) -> bool:
        """Determine if the uri of a service is a service we should forward to."""
        endpoint = service.service_endpoint.uri
        found_forwardable_service = await resolver.is_resolvable(endpoint)
        return found_forwardable_service

    def _create_forward_message(
        self, to: str, next_target: str, message: bytes
    ) -> Dict[Any, Any]:
        return {
            "typ": "application/didcomm-plain+json",
            "type": "https://didcomm.org/routing/2.0/forward",
            "id": str(uuid.uuid4()),
            "to": [to],
            # "expires_time": 123456, #  time to expire the forward message, in epoch time
            "body": {"next": next_target},
            "attachments": [
                {
                    "id": str(uuid.uuid4()),
                    "media_type": "application/didcomm-encrypted+json",
                    "data": {
                        "json": json.loads(message),
                    },
                },
            ],
        }

    async def prepare_forward(
        self,
        crypto: CryptoService[P, S],
        packaging: PackagingService,
        resolver: DIDResolver,
        secrets: SecretsManager[S],
        to: str,
        encoded_message: bytes,
    ) -> Tuple[bytes, DIDCommV2Service]:
        """Prepare a forward message, if necessary.

        Args:
            crypto (CryptoService[P, S]): Crypto service
            packaging (PackagingService): Packaging service
            resolver (DIDResolver): Resolver instance
            secrets (SecretsManager[S]): Secrets manager
            to (str): The recipient of the message. This will be a DID.
            encoded_message (bytes): The encoded message.

        Returns:
            The encoded message, and the services to forward to.
        """

        # Get the initial service
        services = await self._resolve_services(resolver, to)
        chain = [
            {
                "did": to,
                "service": services,
            }
        ]

        # Loop through service DIDs until we run out of DIDs to forward to
        to_did = services[0].service_endpoint.uri
        found_forwardable_service = await self.is_forwardable_service(
            resolver, services[0]
        )
        while found_forwardable_service:
            services = await self._resolve_services(resolver, to_did)
            if services:
                chain.append(
                    {
                        "did": to_did,
                        "service": services,
                    }
                )
                to_did = services[0].service_endpoint.uri
            found_forwardable_service = (
                await self.is_forwardable_service(resolver, services[0])
                if services
                else False
            )

        if not chain[-1]["service"]:
            raise RoutingServiceError(f"No DIDCommV2 service endpoint found for {to}")

        # If we didn't find any services to forward to, just bail
        if len(chain) == 1:
            return (encoded_message, chain[-1]["service"])

        # Grab our target to pack the initial message to, then pack the message
        # for the DID target
        final_destination = chain.pop(0)
        next_target = final_destination["did"]
        packed_message = encoded_message

        # Loop through the entire services chain and pack the message for each
        # layer of mediators
        for service in chain:
            # https://identity.foundation/didcomm-messaging/spec/#sender-process-to-enable-forwarding
            # Respect routing keys by adding the current DID to the front of
            # the list, then wrapping message following routing key order
            routing_keys = service["service"][0].service_endpoint.routing_keys
            routing_keys.insert(0, service["did"])  # prepend did

            # Pack for each key
            while routing_keys:
                key = routing_keys.pop()  # pop from end of list (reverse order)
                packed_message = await packaging.pack(
                    crypto,
                    resolver,
                    secrets,
                    json.dumps(
                        self._create_forward_message(key, next_target, packed_message)
                    ).encode(),
                    [key],
                )
                next_target = key

        # Return the forward-packed message as well as the last service in the
        # chain, which is the destination of the top-level forward message.
        service = final_destination["service"]
        if len(chain):
            service = chain[-1]["service"]
        return (packed_message, service)



"""CryptoService and SecretsManager interfaces for DIDComm Messaging."""

from abc import ABC, abstractmethod
from typing import Generic, Mapping, Optional, Sequence, TypeVar, Union

from pydid import VerificationMethod

from didcomm_messaging.multiformats import multibase, multicodec


class CryptoServiceError(Exception):
    """Represents an error from a CryptoService."""


class PublicKey(ABC):
    """Key representation for CryptoService."""

    type_to_codec: Mapping[str, str] = {
        "Ed25519VerificationKey2018": "ed25519-pub",
        "X25519KeyAgreementKey2019": "x25519-pub",
        "Ed25519VerificationKey2020": "ed25519-pub",
        "X25519KeyAgreementKey2020": "x25519-pub",
    }

    @classmethod
    @abstractmethod
    def from_verification_method(cls, vm: VerificationMethod) -> "PublicKey":
        """Create a Key instance from a DID Document Verification Method."""

    @classmethod
    def key_bytes_from_verification_method(cls, vm: VerificationMethod) -> bytes:
        """Get the key bytes from a DID Document Verification Method."""
        if vm.public_key_multibase and vm.public_key_base58:
            raise ValueError(
                "Only one of public_key_multibase or public_key_base58 must be given"
            )
        if not vm.public_key_multibase and not vm.public_key_base58:
            raise ValueError(
                "One of public_key_multibase or public_key_base58 must be given)"
            )

        if vm.public_key_multibase:
            decoded = multibase.decode(vm.public_key_multibase)
            if len(decoded) == 32:
                # No multicodec prefix
                return decoded
            else:
                codec, decoded = multicodec.unwrap(decoded)
                expected_codec = cls.type_to_codec.get(vm.type)
                if not expected_codec:
                    raise ValueError("Unsupported verification method type")
                if codec.name != expected_codec:
                    raise ValueError("Type and codec mismatch")
                return decoded

        if vm.public_key_base58:
            return multibase.decode("z" + vm.public_key_base58)

        raise ValueError("Invalid verification method")

    @property
    @abstractmethod
    def kid(self) -> str:
        """Get the key ID."""

    @property
    @abstractmethod
    def multikey(self) -> str:
        """Get the key in multikey format."""


class SecretKey(ABC):
    """Secret Key Type."""

    @property
    @abstractmethod
    def kid(self) -> str:
        """Get the key ID."""


P = TypeVar("P", bound=PublicKey)
S = TypeVar("S", bound=SecretKey)


class CryptoService(ABC, Generic[P, S]):
    """Key Management Service (CryptoService) interface for DIDComm Messaging."""

    @abstractmethod
    async def ecdh_es_encrypt(self, to_keys: Sequence[P], message: bytes) -> bytes:
        """Encode a message into DIDComm v2 anonymous encryption."""

    @abstractmethod
    async def ecdh_es_decrypt(self, wrapper: Union[str, bytes], recip_key: S) -> bytes:
        """Decode a message from DIDComm v2 anonymous encryption."""

    @abstractmethod
    async def ecdh_1pu_encrypt(
        self,
        to_keys: Sequence[P],
        sender_key: S,
        message: bytes,
    ) -> bytes:
        """Encode a message into DIDComm v2 authenticated encryption."""

    @abstractmethod
    async def ecdh_1pu_decrypt(
        self,
        wrapper: Union[str, bytes],
        recip_key: S,
        sender_key: P,
    ) -> bytes:
        """Decode a message from DIDComm v2 authenticated encryption."""

    @classmethod
    @abstractmethod
    def verification_method_to_public_key(cls, vm: VerificationMethod) -> P:
        """Convert a verification method to a public key."""


class SecretsManager(ABC, Generic[S]):
    """Secrets Resolver interface.

    Thie secrets resolver may be used to supplement the CryptoService backend to provide
    greater flexibility.
    """

    @abstractmethod
    async def get_secret_by_kid(self, kid: str) -> Optional[S]:
        """Get a secret key by its ID."""



"""Test the authlib crypto service implementation."""

from authlib.jose import OKPKey
import pytest

from didcomm_messaging.crypto.backend.authlib import (
    AuthlibCryptoService,
    AuthlibKey,
    AuthlibSecretKey,
)


ALICE_KID = "did:example:alice#key-1"
BOB_KID = "did:example:bob#key-1"
CAROL_KID = "did:example:carol#key-2"
MESSAGE = b"Expecto patronum"


@pytest.fixture
def crypto():
    yield AuthlibCryptoService()


def test_multikey_roundtrip():
    """Test multikey round trip."""
    key = OKPKey.generate_key("Ed25519", is_private=False)
    multikey = AuthlibKey.key_to_multikey(key)
    back_again = AuthlibKey.multikey_to_key(multikey)
    assert key.as_dict() == back_again.as_dict()


@pytest.mark.asyncio
async def test_1pu_round_trip(crypto: AuthlibCryptoService):
    """Test 1PU round trip."""
    alice_sk = OKPKey.generate_key("X25519", is_private=True)
    alice_pk = alice_sk.get_public_key()
    bob_sk = OKPKey.generate_key("X25519", is_private=True)
    bob_pk = bob_sk.get_public_key()

    bob_key = AuthlibKey(bob_sk, BOB_KID)
    bob_priv_key = AuthlibSecretKey(bob_sk, BOB_KID)

    alice_key = AuthlibKey(alice_sk, ALICE_KID)
    alice_priv_key = AuthlibSecretKey(alice_sk, ALICE_KID)

    enc_message = await crypto.ecdh_1pu_encrypt([bob_key], alice_priv_key, MESSAGE)

    plaintext = await crypto.ecdh_1pu_decrypt(enc_message, bob_priv_key, alice_key)
    assert plaintext == MESSAGE


@pytest.mark.asyncio
async def test_es_round_trip(crypto: AuthlibCryptoService):
    """Test ECDH-ES round trip."""
    alice_sk = OKPKey.generate_key("X25519", is_private=True)
    alice_pk = alice_sk.get_public_key()
    bob_sk = OKPKey.generate_key("X25519", is_private=True)
    bob_pk = bob_sk.get_public_key()

    bob_key = AuthlibKey(bob_sk, BOB_KID)
    bob_priv_key = AuthlibSecretKey(bob_sk, BOB_KID)

    alice_key = AuthlibKey(alice_sk, ALICE_KID)
    alice_priv_key = AuthlibSecretKey(alice_sk, ALICE_KID)

    enc_message = await crypto.ecdh_es_encrypt([bob_key], MESSAGE)

    plaintext = await crypto.ecdh_es_decrypt(enc_message, bob_priv_key)
    assert plaintext == MESSAGE



"""Quickstart helpers for beginner users of DIDComm."""

from typing import (
    Optional,
    Dict,
    List,
    Any,
    Union,
    Callable,
    Awaitable,
    Tuple,
)
import aiohttp
import json
import logging
import uuid

from did_peer_2 import KeySpec, generate
from pydid.did import DID

from aries_askar import Key, KeyAlg
from didcomm_messaging import DIDCommMessaging
from didcomm_messaging.crypto.backend.askar import AskarCryptoService, AskarSecretKey
from didcomm_messaging.crypto.backend.basic import InMemorySecretsManager
from didcomm_messaging.multiformats import multibase, multicodec
from didcomm_messaging.packaging import PackagingService
from didcomm_messaging.resolver import PrefixResolver
from didcomm_messaging.resolver.web import DIDWeb
from didcomm_messaging.resolver.peer import Peer2, Peer4
from didcomm_messaging.routing import RoutingService

JSON_OBJ = Dict[str, Any]
Attachment = JSON_OBJ
JSON_VALUE = Union[None, str, int, bool, float, JSON_OBJ, List[Any]]

LOG = logging.getLogger(__name__)


def generate_did() -> Tuple[DID, Tuple[Key, Key]]:
    """Use Askar to generate encryption/verification keys, then return a DID from both."""

    verkey = Key.generate(KeyAlg.ED25519)
    xkey = Key.generate(KeyAlg.X25519)
    did = generate(
        [
            KeySpec.verification(
                multibase.encode(
                    multicodec.wrap("ed25519-pub", verkey.get_public_bytes()),
                    "base58btc",
                )
            ),
            KeySpec.key_agreement(
                multibase.encode(
                    multicodec.wrap("x25519-pub", xkey.get_public_bytes()), "base58btc"
                )
            ),
        ],
        [
            {
                "type": "DIDCommMessaging",
                "serviceEndpoint": {
                    "uri": "didcomm:transport/queue",
                    "accept": ["didcomm/v2"],
                    "routingKeys": [],
                },
            }
        ],
    )
    return did, (verkey, xkey)


async def setup_default(did: DID, did_secrets: Tuple[Key, Key]) -> DIDCommMessaging:
    """Setup a pre-configured DIDCommMessaging instance."""

    # The Crypto Service is used to encrypt, decrypt, sign and verify messages.
    # Askar is a pretty solid choice for these tasks.
    crypto = AskarCryptoService()

    # The secrets manager is used to hold the private encryption/signing keys
    # to a DID and provide a mapping from the public keys to the private keys.
    # The keys are primarily used by the crypto service when the packaging
    # service is packing/unpacking a signed/encrypted message
    secrets = InMemorySecretsManager()

    # The Resolver is used to turn a DID into a valid DID Document.
    # DIDDocuments contain the public keys of the DID, as well as provide an
    # endpoint in which a message can be sent to. DIDDocuments can refer to a
    # DID as their endpoint, in which case, that DID will also be unwrapped
    # into it's own DIDDocument.
    #
    # At present, the PrefixResolver is used to determine which library should
    # be used to convert a DID into a DIDDocument.
    resolver = PrefixResolver(
        {
            "did:peer:2": Peer2(),
            "did:peer:4": Peer4(),
            "did:web:": DIDWeb(),
        }
    )

    # The Packaging Service is where a lot of the magic happens. Similar to a
    # shipping box, the PackagingService will "pack" and "unpack" an encrypted
    # message. When packing a message, the PackagingService will encrypt the
    # message to a single target, however. If the message needs to be forwarded
    # (because the recipient is behind a relay), then those messages will need
    # to be handled by the RoutingService.
    packer = PackagingService()

    # The RoutingService handles the routing of messages through relays. When a
    # message needs to be forwarded, the RoutingService will handle wrapping
    # each encrypted message within a forward message. The built-in
    # RoutingService allows for multiple levels of relays, so you don't need to
    # worry about how a message is routed to the recipient.
    router = RoutingService()

    # Once everything is setup, we need to store the DID Secrets within the
    # SecretsService. We do this by taking the secrets that were passed in,
    # converting them to AskarSecretKey objects (since we are using the
    # AskarCryptoService), then we associate them with the public key lookup
    # value that will be sent to us.
    verkey, xkey = did_secrets
    await secrets.add_secret(AskarSecretKey(verkey, f"{did}#key-1"))
    await secrets.add_secret(AskarSecretKey(xkey, f"{did}#key-2"))

    # Finally, we put it all together in the DIDCommMessaging class. The
    # DIDCommMessaging handles the orchestration of each individual service,
    # ensuring that messages get packed and delivered via a simple and straight
    # forward interface.
    DMP = DIDCommMessaging(
        crypto=crypto,
        secrets=secrets,
        resolver=resolver,
        packaging=packer,
        routing=router,
    )

    return DMP


async def send_http_message(
    dmp: DIDCommMessaging, my_did: DID, message: Dict[str, Any], target: DID
) -> Optional[Dict[str, Any]]:
    """Send a message via HTTP."""

    # Ensure an ID is on the message
    if "id" not in message or not message["id"]:
        message["id"] = str(uuid.uuid4())

    # Ensure that a typ is on the message
    if "typ" not in message or not message["typ"]:
        message["typ"] = "application/didcomm-plain+json"

    # Ensure that a return-route is on the message
    if "return_route" not in message or not message["return_route"]:
        message["return_route"] = "all"

    # Encrypt/pack the message to our target
    packy = await dmp.pack(
        message=message,
        to=target,
        frm=my_did,
    )
    packed = packy.message

    # Get the first http endpoint from the last DID in the DID chain
    endpoint = packy.get_endpoint("http")

    async with aiohttp.ClientSession() as session:
        LOG.info("posting message type %s to %s", message["type"], endpoint)

        async with session.post(endpoint, data=packed) as resp:
            # Get the message from the response and prepare for decryption
            packed = await resp.text()

            # Dump useful information about the message
            LOG.debug("posted message: %s", message)
            LOG.debug("message ID: %s", message["id"])
            LOG.debug("response code: %s", resp.status)
            LOG.debug("response message: %s", packed)

            # Raise an exception if the destination did not return success
            if resp.status != 200:
                raise Exception(
                    "Destination responded with error: code=%s message=%s"
                    % (resp.status, packed)
                )

            # If the HTTP enpoint responded with a message, decode it
            if len(packed) > 0:
                unpacked = await dmp.unpack(packed)
                msg = unpacked.message
                LOG.debug("Raw message from remote %s", msg)
                return msg
    return


async def setup_relay(
    dmp: DIDCommMessaging, my_did: DID, relay_did: DID, verkey: Key, xkey: Key
) -> Union[DID, None]:
    """Negotiate services with an inbound relay.

    Returns a DID upon successful negotiation.
    """

    # Request mediation from the inbound relay
    message = {
        "type": "https://didcomm.org/coordinate-mediation/3.0/mediate-request",
        "id": str(uuid.uuid4()),
        "body": {},
        "from": my_did,
        "to": [relay_did],
    }
    message = await send_http_message(dmp, my_did, message, target=relay_did)

    # Verify that mediation-access has been granted
    if message["type"] == "https://didcomm.org/coordinate-mediation/3.0/mediate-deny":
        return
    if message["type"] != "https://didcomm.org/coordinate-mediation/3.0/mediate-grant":
        # We shouldn't run into this case, but it's possible
        raise Exception("Unknown response type received: %s" % message["type"])

    # Create a new DID with an updated service endpoint, pointing to our relay
    relay_did = message["body"]["routing_did"][0]
    new_did = generate(
        [
            KeySpec.verification(
                multibase.encode(
                    multicodec.wrap("ed25519-pub", verkey.get_public_bytes()),
                    "base58btc",
                )
            ),
            KeySpec.key_agreement(
                multibase.encode(
                    multicodec.wrap("x25519-pub", xkey.get_public_bytes()), "base58btc"
                )
            ),
        ],
        [
            {
                "type": "DIDCommMessaging",
                "serviceEndpoint": {
                    "uri": relay_did,
                    "accept": ["didcomm/v2"],
                },
            }
        ],
    )
    LOG.info("relayed did: %s", new_did)

    # A couple of helpers variables to simplify the next few lines
    resolver = dmp.resolver
    secrets = dmp.secrets

    # Add the DID to our secrets manager so that we can unpack messages
    # destined to us via our new DID
    doc = await resolver.resolve_and_parse(new_did)
    # New format, key-# is in order of keys in did
    await secrets.add_secret(AskarSecretKey(verkey, f"{new_did}#key-1"))
    await secrets.add_secret(AskarSecretKey(xkey, f"{new_did}#key-2"))

    # Legacy formats
    # verkey
    await secrets.add_secret(AskarSecretKey(verkey, doc.authentication[0]))
    # xkey
    await secrets.add_secret(AskarSecretKey(xkey, doc.key_agreement[0]))

    # Send a message to the relay informing it of our new endpoint that people
    # should contact us by
    message = {
        "type": "https://didcomm.org/coordinate-mediation/3.0/recipient-update",
        "id": str(uuid.uuid4()),
        "body": {
            "updates": [
                {
                    "recipient_did": new_did,
                    "action": "add",
                },
            ],
        },
        "from": my_did,
        "to": [relay_did],
    }
    message = await send_http_message(dmp, my_did, message, target=relay_did)

    return new_did


async def _message_callback(msg: Dict[str, Any]) -> None:
    if msg["type"] == "https://didcomm.org/basicmessage/2.0/message":
        logmsg = msg.body["content"].replace("\n", " ").replace("\r", "")
        LOG.info("Got message: %s", logmsg)


async def fetch_relayed_messages(
    dmp: DIDCommMessaging,
    my_did: DID,
    relay_did: DID,
    callback: Callable[[Dict[str, Any]], Awaitable[None]] = _message_callback,
) -> None:
    """Fetch stored messages from the relay."""

    # Fetch a count of all stored messages
    message = {
        "type": "https://didcomm.org/messagepickup/3.0/status-request",
        "id": str(uuid.uuid4()),
        "body": {},
        "from": my_did,
        "to": [relay_did],
    }
    message = await send_http_message(dmp, my_did, message, target=relay_did)

    # If there's no messages, we can return early
    if message["body"]["message_count"] == 0:
        return []

    # Request messages that are stored at the relay, according to the
    # message_count returned in the previous call
    message = {
        "type": "https://didcomm.org/messagepickup/3.0/delivery-request",
        "id": str(uuid.uuid4()),
        "body": {
            "limit": message["body"]["message_count"],
        },
        "from": my_did,
        "to": [relay_did],
    }
    message = await send_http_message(dmp, my_did, message, target=relay_did)

    # Handle each stored message is order we receive it
    for attach in message["attachments"]:
        LOG.info("Received message %s", attach["id"][:-58])

        # Decrypt/Unpack the encrypted message attachment
        unpacked = await dmp.unpack(json.dumps(attach["data"]["json"]))
        msg = unpacked.message

        # Call callback if it exists, passing the message in
        if callback:
            await callback(msg)

    message = {
        "type": "https://didcomm.org/messagepickup/3.0/messages-received",
        "id": str(uuid.uuid4()),
        "body": {
            "message_id_list": [msg["id"] for msg in message["attachments"]],
        },
        "from": my_did,
        "to": [relay_did],
    }
    await send_http_message(dmp, my_did, message, target=relay_did)


# '##:::::'##:'########:'########:::'######:::'#######:::'######::'##:::'##:
#  ##:'##: ##: ##.....:: ##.... ##:'##... ##:'##.... ##:'##... ##: ##::'##::
#  ##: ##: ##: ##::::::: ##:::: ##: ##:::..:: ##:::: ##: ##:::..:: ##:'##:::
#  ##: ##: ##: ######::: ########::. ######:: ##:::: ##: ##::::::: #####::::
#  ##: ##: ##: ##...:::: ##.... ##::..... ##: ##:::: ##: ##::::::: ##. ##:::
#  ##: ##: ##: ##::::::: ##:::: ##:'##::: ##: ##:::: ##: ##::: ##: ##:. ##::
# . ###. ###:: ########: ########::. ######::. #######::. ######:: ##::. ##:
# :...::...:::........::........::::......::::.......::::......:::..::::..::

# The following functions are used for establishing a persistant WebSocket
# connection with a mediator/relay. This allows messages to be delivered as
# soon as they arrive instead of waiting to be picked up with the pickup
# protocol.

# Because the quickstart can be used without using websockets or connecting to
# an inbound relay/mediator. If we don't have the necessary libraries, we can
# just raise a warning about the lack of support.
try:
    import websockets
    import asyncio
except ImportError:
    import warnings

    warnings.warn(
        "Missing websockets or asyncio import, live-delivery support will be unavailable",
        ImportWarning,
    )


async def handle_websocket(
    dmp: DIDCommMessaging,
    relay_did: DID,
    mediator_websocket: Awaitable["websockets.WebSocketClientProtocol"],
    live_delivery_message: bytes,
    callback: Callable[[Dict[str, Any]], Awaitable[None]] = _message_callback,
):
    """Loop over messages received and process them."""

    # Using async with on the websocket allows us to wait for then loop over
    # new messages that come in over the websocket.
    async with mediator_websocket as websocket:
        # Send our live-delivery request to the inbound relay.
        await websocket.send(live_delivery_message)
        LOG.debug("Connected to WebSocket and requested Live Delivery.")

        # Loop over all incoming messages
        while True:
            message = await websocket.recv()
            LOG.debug("Received message over websocket")

            try:
                # Unpack/Decrypt the message, decode it, and load the JSON into
                # a native python object.
                unpacked_message = await dmp.unpack(message)
                msg = unpacked_message.message
                LOG.debug("Received websocket message %s", msg["type"])

                # If the message is not from the relay, process it via the callback
                if msg["from"] != relay_did:
                    await callback(msg)

                # NOTE: While out of scope for this example, when implementing
                # your own handler, it's advisable to handle messages that come
                # from the relay. For example: the relay will respond to the
                # live-delivery call with a pickup protocol status message. The
                # status message may contain a message count that is greater
                # than zero. If there are messages that are waiting, it is
                # advisable to request the messages and process them.

            except Exception as err:
                LOG.error("Error encountered while decrypting websocket message")
                LOG.exception(err)

        # Clean up the websocket
        await websocket.close()


async def activate_websocket(
    dmp: DIDCommMessaging,
    my_did: DID,
    relay_did: DID,
    callback: Callable[[Dict[str, Any]], Awaitable[None]] = _message_callback,
    create_task: bool = False,
) -> Union[Awaitable[None], "asyncio.Task"]:
    """Connect to a websocket and request message forwarding."""

    # Construct a message to request live-delivery and pack it
    message = {
        "type": "https://didcomm.org/messagepickup/3.0/live-delivery-change",
        "id": str(uuid.uuid4()),
        "body": {
            "live_delivery": True,
        },
        "from": my_did,
        "to": [relay_did],
    }
    packed = await dmp.pack(
        message=message,
        to=relay_did,
        frm=my_did,
    )

    # Get the websocket endpoint from the packed message. This will pull out
    # the endpoint that is closest to us, the sender.
    endpoint = packed.get_endpoint("ws")
    LOG.debug("Relay Websocket Address: %s", endpoint)

    if endpoint:
        # Connect to the websocket and prepare our handler function
        mediator_websocket = websockets.connect(uri=endpoint)
        websocket_handler = handle_websocket(
            dmp,
            relay_did,
            mediator_websocket,
            packed.message,
            callback,
        )

        # Create an asyncio task upon request to run in another async "thread",
        # otherwise we'll just return the function that runs in the thread.
        if create_task:
            return asyncio.create_task(websocket_handler)
        else:
            return websocket_handler


async def websocket_loop(
    dmp: DIDCommMessaging,
    my_did: DID,
    relay_did: DID,
    callback: Callable[[Dict[str, Any]], Awaitable[None]] = _message_callback,
) -> None:
    """Run the websocket handler in a task and reconnect on failure."""

    # Helper method to create the async task
    async def create_task():
        return await activate_websocket(dmp, my_did, relay_did, callback, True)

    # Create the initial task
    mediator_websocket_proc = await create_task()

    # Check to make sure the thread is still running every 5 seconds (to give
    # the OS some breathing room). Should the thread "crash" or exit for any
    # reason, log any exceptions and attempt to restart it.
    while True:
        await asyncio.sleep(5)
        if mediator_websocket_proc.done():
            LOG.exception(mediator_websocket_proc.exception())
            LOG.error("Websocket died, re-establishing connection!")
            mediator_websocket_proc = await create_task()
